import pytest

from pyle38 import Tile38
from pyle38.errors import Tile38KeyNotFoundError

from .helper.random_data import random_latitude, random_longitude, random_string


@pytest.mark.asyncio
async def test_command_rename(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    newkey = random_string()
    lat = random_latitude()
    lon = random_longitude()

    response = await tile38.set(key, oid).point(lat, lon).exec()
    assert response.ok

    response = await tile38.rename(key, newkey)
    assert response.ok

    with pytest.raises(Tile38KeyNotFoundError):
        await tile38.get(key, oid).asObject()


@pytest.mark.asyncio
async def test_command_renamex(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    newkey = random_string()
    lat = random_latitude()
    lon = random_longitude()

    response = await tile38.set(key, oid).point(lat, lon).exec()
    assert response.ok

    response = await tile38.set(newkey, oid).point(lat, lon).exec()
    assert response.ok

    response = await tile38.rename(key, newkey, nx=True)
    assert response.ok

    response = await tile38.get(key, oid).asObject()
    assert response.ok

    response = await tile38.get(newkey, oid).asObject()
    assert response.ok



from typing import Any, ClassVar, Generic, Literal, TypeVar

from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict

T = TypeVar("T")
S = TypeVar("S", bound=str)


class BaseModel(PydanticBaseModel):
    def dict(self, exclude_unset: bool = True, **kwargs: Any) -> dict[str, Any]:
        return super().model_dump(exclude_unset=exclude_unset, **kwargs)  # type: ignore[no-any-return,unused-ignore]


class LatLon(BaseModel):
    lat: float
    lon: float
    z: float | None = None


class NeSw(BaseModel):
    ne: LatLon
    sw: LatLon


Fields = dict[str, Any]  # pyright: ignore[reportExplicitAny]
Meta = dict[str, str]


class WithFieldsArray:
    fields: list[Any] | None = None  # pyright: ignore[reportExplicitAny]


class JSONResponse(BaseModel):
    ok: bool
    elapsed: str
    err: str | None = None


class ExistsResponse(JSONResponse, BaseModel):
    exists: bool


class Object(BaseModel, WithFieldsArray, Generic[T]):
    object: T
    id: str | int
    distance: float | None = None


class ObjectResponse(JSONResponse, BaseModel, Generic[T]):
    object: T
    fields: Fields | None = None


class ObjectsResponse(JSONResponse, BaseModel, WithFieldsArray, Generic[T]):
    objects: list[Object[T]] = []
    count: int
    cursor: int


class IdsResponse(JSONResponse):
    ids: list[str]
    count: int
    cursor: int


class CountResponse(JSONResponse):
    count: int
    cursor: int


class PointResponse(JSONResponse):
    point: LatLon
    fields: Fields | None = None


class Point(BaseModel, WithFieldsArray):
    point: LatLon
    id: str | int
    distance: int | None = None


class PointsResponse(JSONResponse, WithFieldsArray):
    points: list[Point] = []
    count: int
    cursor: int


class Hash(BaseModel, WithFieldsArray):
    hash: str
    id: str | int
    distance: float | None = None


class HashResponse(JSONResponse):
    hash: str
    fields: Fields | None = None


class HashesResponse(JSONResponse, WithFieldsArray):
    hashes: list[Hash]
    count: int
    cursor: int


class Bounds(BaseModel, WithFieldsArray):
    bounds: NeSw
    id: str | int
    distance: float | None = None


class BoundsNeSwResponse(JSONResponse):
    bounds: NeSw
    fields: Fields | None = None


class BoundsNeSwResponses(JSONResponse, WithFieldsArray):
    bounds: list[Bounds]
    count: int
    cursor: int


Position = list[float]


class Polygon(BaseModel):
    type: Literal["Polygon"] = "Polygon"
    coordinates: list[list[Position]]


class BoundsResponse(JSONResponse):
    bounds: Polygon


class KeysResponse(JSONResponse):
    keys: list[str]


class TTLResponse(JSONResponse):
    # if no ttl -1
    ttl: float


class Stats(BaseModel):
    in_memory_size: int | float
    num_objects: int | float
    num_points: int | float
    num_strings: int | float


class StatsResponse(JSONResponse):
    stats: list[Stats] | None = []


class ServerStatsLeader(BaseModel):
    aof_size: int | float
    avg_item_size: int | float
    cpus: int | float
    heap_released: int | float
    heap_size: int | float
    http_transport: bool
    id: str
    in_memory_size: int | float
    max_heap_size: int | float
    mem_alloc: int | float
    num_collections: int
    num_hooks: int
    num_objects: int
    num_points: int
    num_strings: int
    pid: int
    pointer_size: int | float
    read_only: bool
    threads: int
    version: str


class ServerStatsFollower(ServerStatsLeader):
    caught_up: bool
    caught_up_once: bool
    following: str


class ServerStatsResponseLeader(JSONResponse):
    stats: ServerStatsLeader


class ServerStatsResponseFollower(JSONResponse):
    stats: ServerStatsFollower


class ServerStatsExtended(BaseModel):
    alloc_bytes: int
    alloc_bytes_total: int
    buck_hash_sys_bytes: int
    frees_total: float
    gc_cpu_fraction: float
    gc_sys_bytes: float
    go_goroutines: int
    go_threads: int
    go_version: str
    heap_alloc_bytes: int
    heap_idle_bytes: int
    heap_inuse_bytes: int
    heap_objects: int
    heap_released_bytes: int
    heap_sys_bytes: int
    last_gc_time_seconds: float
    lookups_total: int
    mallocs_total: int
    mcache_inuse_bytes: int
    mcache_sys_bytes: int
    mspan_inuse_bytes: int
    mspan_sys_bytes: int
    next_gc_bytes: int
    other_sys_bytes: int
    stack_inuse_bytes: int
    stack_sys_bytes: int
    sys_bytes: int
    sys_cpus: int
    tile38_aof_current_rewrite_time_sec: int
    tile38_aof_enabled: bool
    tile38_aof_last_rewrite_time_sec: int
    tile38_aof_rewrite_in_progress: bool
    tile38_aof_size: int
    tile38_avg_point_size: int
    tile38_cluster_enabled: bool
    tile38_connected_clients: int
    tile38_connected_slaves: int
    tile38_expired_keys: int
    tile38_http_transport: bool
    tile38_id: str
    tile38_in_memory_size: int
    tile38_max_heap_size: int
    tile38_num_collections: int
    tile38_num_hooks: int
    tile38_num_objects: int
    tile38_num_points: int
    tile38_num_strings: int
    tile38_num_object_groups: int
    tile38_num_hook_groups: int
    tile38_pid: int
    tile38_pointer_size: int
    tile38_read_only: bool
    tile38_total_commands_processed: int
    tile38_total_connections_received: int
    tile38_total_messages_sent: int
    tile38_type: str
    tile38_uptime_in_seconds: float
    tile38_version: str


class ServerStatsExtendedResponse(JSONResponse):
    stats: ServerStatsExtended


ConfigKeys = Literal[
    "requirepass", "leaderauth", "protected-mode", "maxmemory", "autogc", "keepalive"
]


class ConfigGetResponse(JSONResponse):
    properties: dict[ConfigKeys, str]


class JSONGetResponse(JSONResponse):
    value: int | float | str | None = "{}"


class Hooks(BaseModel):
    name: str
    endpoints: list[str]
    key: str
    meta: dict[str, Any]  # pyright: ignore[reportExplicitAny]
    command: list[str]


class HooksResponse(JSONResponse):
    hooks: list[Hooks]


class Chans(BaseModel):
    name: str
    key: str
    meta: dict[str, Any]  # pyright: ignore[reportExplicitAny]
    command: list[str]


class ChansResponse(JSONResponse):
    chans: list[Chans]


class PingResponse(JSONResponse):
    ping: Literal["pong"] = "pong"


FenceDetect = Literal["enter", "exit", "inside", "outside", "crosses"]
FenceCommand = Literal["set", "del"]


class GeoFence(BaseModel, Generic[T]):
    command: FenceCommand
    group: str
    detect: FenceDetect
    hook: str
    key: str
    time: str
    id: str
    object: T


class Info(BaseModel):
    aof_current_rewrite_time_sec: int
    aof_enabled: int
    aof_last_rewrite_time_sec: int
    aof_rewrite_in_progress: int
    cluster_enabled: int
    connected_clients: int
    connected_slaves: int
    expired_keys: int
    redis_version: str
    role: Literal["master", "slave"]
    tile38_version: str
    total_messages_sent: int
    total_connections_received: int
    total_commands_processed: int
    uptime_in_seconds: int
    used_cpu_sys: int
    used_cpu_sys_children: int
    used_cpu_user: int
    used_cpu_user_children: int
    used_memory: int


class InfoFollower(Info):
    master_host: str
    master_port: int


class InfoLeader(Info):
    # to allow for additional slaves
    # slave0, slave1..
    model_config: ClassVar[ConfigDict] = ConfigDict(extra="allow")


class InfoFollowerResponse(JSONResponse):
    info: InfoFollower


class InfoLeaderResponse(JSONResponse):
    info: InfoLeader



from __future__ import annotations

from collections.abc import Sequence
from typing import Literal

from ..client import Client, Command, CommandArg, CommandArgs
from ..models import Options
from ..responses import CountResponse, IdsResponse, ObjectsResponse
from .executable import Compiled, Executable
from .whereable import Whereable

Format = Literal["OBJECTS", "COUNT", "IDS"]
Output = Sequence[Format | int]


class Search(Executable, Whereable):
    _key: str
    _command: Literal["SEARCH"]
    _options: Options
    _output: Output | None = None
    _all: bool = False

    def __init__(self, client: Client, key: str) -> None:
        super().__init__(client)

        self.key(key)
        self._options = {}
        self._where = []
        self._wherein = []

    def key(self, key: str) -> Search:
        self._key = key

        return self

    def cursor(self, value: int) -> Search:
        self._options["cursor"] = value

        return self

    def limit(self, value: int) -> Search:
        self._options["limit"] = value

        return self

    def match(self, value: str) -> Search:
        self._options["match"] = value

        return self

    def asc(self, flag: bool = True) -> Search:
        self._options["asc"] = flag

        if flag:
            self._options["desc"] = False

        return self

    def desc(self, flag: bool = True) -> Search:
        self._options["desc"] = flag

        if flag:
            self._options["asc"] = False

        return self

    def output(self, fmt: Format) -> Search:
        if fmt == "OBJECTS":
            self._output = None
        elif fmt == "COUNT" or fmt == "IDS":
            self._output = [fmt]
        return self

    async def asCount(self) -> CountResponse:
        self.output("COUNT")

        return CountResponse(**(await self.exec()))

    async def asIds(self) -> IdsResponse:
        self.output("IDS")

        return IdsResponse(**(await self.exec()))

    async def asStringObjects(self) -> ObjectsResponse[str]:
        self.output("OBJECTS")

        return ObjectsResponse[str](**(await self.exec()))

    def __compile_options(self) -> CommandArgs:
        commands: list[CommandArg] = []

        # raises mypy: TypedDict key must be string literal
        # open PR: https://github.com/python/mypy/issues/7867
        for k in self._options:
            if isinstance(self._options[k], bool):  # type: ignore
                if self._options[k]:  # type: ignore
                    commands.append(k.upper())
            elif self._options[k] or self._options[k] == 0:  # type: ignore
                commands.extend([k.upper(), self._options[k]])  # type: ignore

        return commands

    def compile(self) -> Compiled:
        return [
            Command.SEARCH.value,
            [
                self._key,
                *(self.__compile_options()),
                *(self.compile_where()),
                *(self.compile_wherein()),
                *(self._output if self._output else []),
            ],
        ]  # type: ignore



import pytest

from pyle38 import Tile38

from .helper.random_data import random_string


@pytest.mark.asyncio
async def test_command_ttl(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    expire_in = 5

    response = await tile38.set(key, oid).ex(expire_in).point(1, 1).exec()
    assert response.ok

    response = await tile38.ttl(key, oid)
    assert response.ok
    assert response.ttl < expire_in



import pytest

from pyle38 import Tile38
from pyle38.commands.fget import FGet
from pyle38.errors import Tile38FieldNotFoundError

from .helper.random_data import random_integer, random_point_feature, random_string


@pytest.mark.asyncio
async def test_command_fget_compile(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    fkey = random_string()

    received = FGet(tile38.client, key, oid, fkey).compile()
    expected = ["FGET", [key, oid, fkey]]

    assert received == expected


@pytest.mark.asyncio
async def test_command_fget(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    fkey = random_string()
    fvalue = random_integer(1, 10)
    fields = {fkey: fvalue}
    obj = random_point_feature()

    response = await tile38.set(key, oid).object(obj).exec()
    assert response.ok

    response = await tile38.fset(key, oid, fields).xx().exec()
    assert response.ok

    response = await tile38.get(key, oid).withfields().asObject()
    assert response.ok
    assert response.fields == fields

    response = await tile38.fget(key, oid, fkey).exec()
    assert response.ok
    assert response.value == fvalue


@pytest.mark.asyncio
async def test_command_fget_raises(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    fkey = random_string()
    fvalue = random_integer(1, 10)
    fields = {fkey: fvalue}
    obj = random_point_feature()

    response = await tile38.set(key, oid).object(obj).exec()
    assert response.ok

    response = await tile38.fset(key, oid, fields).xx().exec()
    assert response.ok

    response = await tile38.get(key, oid).withfields().asObject()
    assert response.ok
    assert response.fields == fields

    with pytest.raises(Tile38FieldNotFoundError):
        response = await tile38.fget(key, oid, "unknown").exec()



import pytest

from pyle38 import Tile38

from .helper.random_data import random_latitude, random_longitude, random_string


@pytest.mark.asyncio
async def test_command_keys(tile38_with_follower: Tile38) -> None:
    tile38 = tile38_with_follower

    key = random_string()
    oid = random_string()

    response = (
        await tile38.set(key, oid).point(random_latitude(), random_longitude()).exec()
    )
    assert response.ok

    response = await tile38.keys()
    assert response.ok
    assert response.keys == [key]



import json

from redis.typing import EncodableT

from .errors import (
    Tile38Error,
    Tile38IdNotFoundError,
    Tile38KeyNotFoundError,
    Tile38NotCaughtUpError,
    Tile38PathNotFoundError,
)


def parse_response(
    response: bytes
    | memoryview
    | str
    | int
    | float
    | list[EncodableT]
    | None
    | None = None,
) -> dict[str, float | str | int | list | dict]:
    if not isinstance(response, str) or isinstance(response, bytes):
        raise Tile38Error("invalid response")  # noqa: TRY003

    try:
        obj = json.loads(response)
    except Exception as e:
        raise Tile38Error(e) from e

    msg = "unknown"

    if not obj["ok"]:
        msg = obj["err"]

        if "key not found" in msg:
            raise Tile38KeyNotFoundError(msg)

        if "id not found" in msg:
            raise Tile38IdNotFoundError(msg)

        if "not caught up" in msg:
            raise Tile38NotCaughtUpError(msg)

        if "path not found" in msg:
            raise Tile38PathNotFoundError(msg)

        raise Tile38Error(msg)

    return obj  # type:ignore[no-any-return]
